@@ -1386,6 +1386,9 @@ i830_tv_detect_type (xf86CrtcPtr    crtc,
 	tv_ctl &= ~TV_TEST_MODE_MASK;
 	tv_ctl |= TV_TEST_MODE_MONITOR_DETECT;
 	tv_dac &= ~TVDAC_SENSE_MASK;
+        tv_dac &= ~DAC_A_MASK;
+        tv_dac &= ~DAC_B_MASK;
+        tv_dac &= ~DAC_C_MASK;
 	tv_dac |= (TVDAC_STATE_CHG_EN |
 		TVDAC_A_SENSE_CTL |
 		TVDAC_B_SENSE_CTL |
